# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import datetime

from flask import Request

from palmer.config import default_config


class APIRequest(Request):
    renderer_class = default_config.RENDERER_CLASS

    def __init__(self, *args, **kwargs):
        self.request_at = datetime.datetime.utcnow()
        super(APIRequest, self).__init__(*args, **kwargs)

    @property
    def form_data(self):
        data = self.form.copy()
        if self.files:
            data.update(self.files)
        return data

    @property
    def renderer(self):
        if not hasattr(self, '_renderer'):
            self._renderer = self.renderer_class()
        return self._renderer

    @renderer.setter
    def renderer(self, renderer):
        self._renderer = renderer



# -*- coding:utf-8 -*-
import importlib

from flask._compat import string_types


def _gettext(s):
    return s

DEFAULTS = {
    'GETTEXT_FUNC': _gettext,
    'AUTHENTICATION_CLASSES': (),
    'PERMISSION_CLASSES': (
        'palmer.permissions.AllowAny',
    ),
    'RENDERER_CLASS': 'palmer.renderers.JSONRenderer'
}


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError as exc:
        format_ = "Could not import '%s' for API setting '%s'. %s."
        msg = format_ % (val, setting_name, exc)
        raise ImportError(msg)


def perform_imports(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if isinstance(val, string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [perform_imports(item, setting_name) for item in val]
    return val


class APIConfig(object):
    def __init__(self, user_config=None):
        self.user_config = user_config or {}

    def get_or_default(self, name):
        default = DEFAULTS[name]
        val = self.user_config.get(name, default)
        return perform_imports(val, name)

    @property
    def GETTEXT_FUNC(self):
        default = DEFAULTS['GETTEXT_FUNC']
        val = self.user_config.get('GETTEXT_FUNC', default)
        return val if callable(val) else perform_imports(val, 'GETTEXT_FUNC')

    @property
    def AUTHENTICATION_CLASSES(self):
        return self.get_or_default('AUTHENTICATION_CLASSES')

    @property
    def PERMISSION_CLASSES(self):
        return self.get_or_default('PERMISSION_CLASSES')

    @property
    def RENDERER_CLASS(self):
        return self.get_or_default('RENDERER_CLASS')

default_config = APIConfig()



# -*- coding:utf-8 -*-
from __future__ import unicode_literals, absolute_import

from flask import __version__ as flask_version


def is_flask_legacy():
    v = flask_version.split(".")
    return int(v[0]) == 0 and int(v[1]) < 11



# -*- coding:utf-8 -*-
from flask.sessions import SessionInterface, SessionMixin
from itsdangerous import URLSafeTimedSerializer, BadSignature
from werkzeug.datastructures import CallbackDict


class ItsdangerousSession(CallbackDict, SessionMixin):

    def __init__(self, initial=None):
        def on_update(self):
            self.modified = True
        CallbackDict.__init__(self, initial, on_update)
        self.modified = False


class ItsdangerousSessionInterface(SessionInterface):
    salt = 'cookie-session'
    session_class = ItsdangerousSession

    def get_serializer(self, app):
        if not app.secret_key:
            return None
        return URLSafeTimedSerializer(app.secret_key,
                                      salt=self.salt)

    def open_session(self, app, request):
        s = self.get_serializer(app)
        if s is None:
            return None
        val = request.cookies.get(app.session_cookie_name)
        if not val:
            return self.session_class()
        max_age = app.permanent_session_lifetime.total_seconds()
        try:
            data = s.loads(val, max_age=max_age)
            return self.session_class(data)
        except BadSignature:
            return self.session_class()

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        if not session:
            if session.modified:
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain)
            return
        expires = self.get_expiration_time(app, session)
        val = self.get_serializer(app).dumps(dict(session))
        response.set_cookie(app.session_cookie_name, val,
                            expires=expires, httponly=True,
                            domain=domain)



# -*- coding:utf-8 -*-
from palmer.utils import http_statuses
from palmer.utils.translation import lazy_gettext as _


class APIException(Exception):
    status_code = http_statuses.HTTP_500_INTERNAL_SERVER_ERROR
    message = _('A server error occurred.')

    def __init__(self, message=None, status_code=None, errors=None):
        if errors is None:
            errors = {}
        if message is not None:
            self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.errors = errors

    def __str__(self):
        return self.message


class NotAuthenticated(APIException):
    status_code = http_statuses.HTTP_401_UNAUTHORIZED
    default_detail = _('Authentication credentials were not provided.')


class AuthenticationFailed(APIException):
    status_code = http_statuses.HTTP_401_UNAUTHORIZED
    message = _('Incorrect authentication credentials.')


class PermissionDenied(APIException):
    status_code = http_statuses.HTTP_403_FORBIDDEN
    message = _('You do not have permission to perform this action.')


class ValidationError(APIException):
    status_code = http_statuses.HTTP_400_BAD_REQUEST
    message = _('Invalid input.')


class NotFound(APIException):
    status_code = http_statuses.HTTP_404_NOT_FOUND
    message = _('Not found.')


class MethodNotAllowed(APIException):
    status_code = http_statuses.HTTP_405_METHOD_NOT_ALLOWED
    message = _('Method not allowed.')



# -*- coding:utf-8 -*-
from flask import current_app, request
from flask.views import MethodView

from palmer import exceptions
from palmer.config import default_config


class APIView(MethodView):
    authentication_classes = None
    permission_classes = None
    renderer_class = None

    @property
    def user(self):
        if not hasattr(self, '_user'):
            self.perform_authenticate()
        return self._user

    @user.setter
    def user(self, user):
        self._user = user

    @property
    def platform(self):
        if not hasattr(self, '_platform'):
            self._platform = request.headers.get('X-Platform', None)
        return self._platform

    @property
    def ip_address(self):
        return request.environ.get('REMOTE_ADDR', None)

    @classmethod
    def _get_authentication_classes(cls):
        if cls.authentication_classes is not None:
            return cls.authentication_classes
        if not current_app:
            return default_config.AUTHENTICATION_CLASSES
        return getattr(current_app.api_config, 'AUTHENTICATION_CLASSES',
                       default_config.AUTHENTICATION_CLASSES)

    @classmethod
    def _get_permission_classes(cls):
        if cls.permission_classes is not None:
            return cls.permission_classes
        if not current_app:
            return default_config.PERMISSION_CLASSES
        return getattr(current_app.api_config, 'PERMISSION_CLASSES',
                       default_config.PERMISSION_CLASSES)

    def _get_renderer_class(self):
        if self.renderer_class is not None:
            return self.renderer_class
        if not current_app:
            return default_config.RENDERER_CLASS
        return getattr(current_app.api_config, 'RENDERER_CLASS', default_config.RENDERER_CLASS)

    def get_authenticators(self):
        return [auth() for auth in self._get_authentication_classes()]

    def get_permissions(self):
        return [permission() for permission in self._get_permission_classes()]

    def perform_authenticate(self):
        for authenticator in self.get_authenticators():
            user = authenticator.authenticate()
            if user:
                self.user = user
                return
        self.user = None

    def check_permissions(self):
        """
        Check if the request should be permitted.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self.get_permissions():
            if not permission.has_permission(self):
                raise exceptions.PermissionDenied(message=getattr(permission, 'message', None))

    def permission_denied(self, message=None):
        """
        If request is not permitted, determine what kind of exception to raise.
        """
        if not self.user:
            raise exceptions.NotAuthenticated()
        raise exceptions.PermissionDenied(message=message)

    def dispatch_request(self, *args, **kwargs):
        meth = getattr(self, request.method.lower(), None)
        if meth is None and request.method == 'HEAD':
            meth = getattr(self, 'get', None)
        assert meth is not None, 'Unimplemented method %r' % request.method

        self.perform_authenticate()
        self.check_permissions()

        pre_process = getattr(self, 'pre_process', None)
        if pre_process is not None:
            pre_process(*args, **kwargs)

        meth_pre_process = getattr(self, '{0}_pre_process'.format(request.method.lower()), None)
        if meth_pre_process is not None:
            meth_pre_process(*args, **kwargs)

        resp = meth(*args, **kwargs)
        request.renderer_class = self._get_renderer_class()
        return resp



# -*- coding:utf-8 -*-
import re
from collections import defaultdict

from flask_swagger import _sanitize, _extract_definitions, _parse_docstring

TYPE_TO_SWAGGER = {
    'int': 'integer',
    'str': 'string'
}


def create_doc(summary, description=None, tags=(), parameters=(), responses=None, **kwargs):
    if description is None:
        description = ''
    else:
        description = '\n'.join([text.strip() for text in description.split('\n')])
    doc = dict(
        summary=summary,
        description=description,
        tags=tags,
        parameters=parameters,
        responses=responses or {},
    )
    doc.update(kwargs)
    return doc


def swagger(app, prefix=None, process_doc=_sanitize,
            from_file_keyword=None, template=None):
    """
    Call this from an @app.route method like this
    @app.route('/spec.json')
    def spec():
       return jsonify(swagger(app))

    We go through all endpoints of the app searching for swagger endpoints
    We provide the minimum required data according to swagger specs
    Callers can and should add and override at will

    Arguments:
    app -- the flask app to inspect

    Keyword arguments:
    process_doc -- text sanitization method, the default simply replaces \n with <br>
    from_file_keyword -- how to specify a file to load doc from
    template -- The spec to start with and update as flask-swagger finds paths.
    """
    output = {
        "swagger": "2.0",
        "info": {
            "version": "0.0.0",
            "title": app.name,
        }
    }
    paths = defaultdict(dict)
    definitions = defaultdict(dict)
    if template is not None:
        output.update(template)
        # check for template provided paths and definitions
        for k, v in output.get('paths', {}).items():
            paths[k] = v
        for k, v in output.get('definitions', {}).items():
            definitions[k] = v
    output["paths"] = paths
    output["definitions"] = definitions

    ignore_verbs = {"HEAD", "OPTIONS"}
    # technically only responses is non-optional
    optional_fields = ['tags', 'consumes', 'produces', 'schemes', 'security',
                       'deprecated', 'operationId', 'externalDocs']

    for rule in app.url_map.iter_rules():
        if prefix and rule.rule[:len(prefix)] != prefix:
            continue
        endpoint = app.view_functions[rule.endpoint]
        methods = dict()
        for verb in rule.methods.difference(ignore_verbs):
            if hasattr(endpoint, 'methods') and verb in endpoint.methods:
                verb = verb.lower()
                methods[verb] = endpoint.view_class
            else:
                methods[verb.lower()] = endpoint
        operations = dict()
        for verb, method in methods.items():
            doc_func = getattr(method, '{0}_doc'.format(verb), None)
            if doc_func is None:
                continue
            swag_options = doc_func()(app)
            path_args = re.findall(r'[^<>]+', rule.rule)[1::2]
            path_params = []
            for arg in path_args:
                type_, name = arg.split(':')
                path_params.append(
                    {
                        'in': 'path',
                        'name': name,
                        'type': TYPE_TO_SWAGGER.get(type_, 'string'),
                        'required': True,
                    }
                )
            swag_options['parameters'] = path_params + list(swag_options.get('parameters', []))
            method.swag_options = swag_options

            if hasattr(method, 'swag_options'):
                swag = method.swag_options
                summary = swag.pop('summary', '')
                description = swag.pop('description', '')
            else:
                summary, description, swag = _parse_docstring(method, process_doc,
                                                              from_file_keyword)
            if swag is not None:  # we only add endpoints with swagger data in the docstrings
                defs = swag.get('definitions', [])
                defs = _extract_definitions(defs)
                params = swag.get('parameters', [])
                defs += _extract_definitions(params)
                responses = swag.get('responses', {})
                responses = {
                    str(key): value
                    for key, value in responses.items()
                    }
                if responses is not None:
                    defs = defs + _extract_definitions(responses.values())
                for definition in defs:
                    def_id = definition.pop('id')
                    if def_id is not None:
                        definitions[def_id].update(definition)
                operation = dict(
                    summary=summary,
                    description=description,
                    responses=responses
                )
                # parameters - swagger ui dislikes empty parameter lists
                if len(params) > 0:
                    operation['parameters'] = params
                # other optionals
                for key in optional_fields:
                    if key in swag:
                        operation[key] = swag.get(key)
                operations[verb] = operation

        if len(operations):
            rule = str(rule)
            for arg in re.findall('(<([^<>]*:)?([^<>]*)>)', rule):
                rule = rule.replace(arg[0], '{%s}' % arg[2])
            paths[rule].update(operations)
    return output



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import datetime
import decimal

import msgpack
from flask.json import dumps as json_dumps


class MessagePackEncoder(object):
    def encode(self, obj):
        if isinstance(obj, datetime.datetime):
            return {'__class__': 'datetime', 'as_str': obj.isoformat()}
        elif isinstance(obj, datetime.date):
            return {'__class__': 'date', 'as_str': obj.isoformat()}
        elif isinstance(obj, datetime.time):
            return {'__class__': 'time', 'as_str': obj.isoformat()}
        elif isinstance(obj, decimal.Decimal):
            return {'__class__': 'decimal', 'as_str': str(obj)}
        else:
            return obj


class BaseRenderer(object):
    """
    All renderers should extend this class, setting the `media_type`
    and `format` attributes, and override the `.render()` method.
    """
    media_type = None
    format = None
    charset = 'utf-8'
    render_style = 'text'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        raise NotImplementedError('Renderer class requires .render() to be implemented')


class JSONRenderer(BaseRenderer):
    """
    Renderer which serializes to JSON.
    """
    media_type = 'application/json'
    format = 'json'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into JSON, returning a bytestring.
        """
        return json_dumps(data, ensure_ascii=False)


class MessagePackRenderer(BaseRenderer):
    """
    Renderer which serializes to MessagePack.
    """
    media_type = 'application/msgpack'
    format = 'msgpack'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into MessagePack, returning a bytes.
        """
        if data is None:
            return ''
        return msgpack.packb(data, default=MessagePackEncoder().encode)



# -*- coding:utf-8 -*-
class BasePermission(object):
    """
    Taken from django-rest-framework
    A base class from which all permission classes should inherit.
    """

    def has_permission(self, view):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True


class AllowAny(BasePermission):
    """
    Allow any access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(self, view):
        return True



# -*- coding:utf-8 -*-
class BaseAuthentication(object):
    """
    All authentication classes should extend BaseAuthentication.
    """

    def authenticate(self):
        """
        Authenticate the request and return a two-tuple of (user, token).
        """
        raise NotImplementedError(".authenticate() must be overridden.")



# -*- coding:utf-8 -*-
from flask import current_app
from speaklater import make_lazy_gettext

from palmer.config import default_config


def _lazy_gettext():
    def _(s):
        config = current_app.api_config if current_app else default_config
        return getattr(config, 'GETTEXT_FUNC')(s)
    return _

lazy_gettext = make_lazy_gettext(_lazy_gettext)



import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def readme():
    try:
        with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
            return f.read()
    except (IOError, OSError):
        return ''


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


package = 'palmer'
install_requires = [
    'flask >= 0.10',
    'flask-swagger >= 0.2.13',
    'inflection >= 0.3.1',
    'msgpack-python >= 0.4.8',
    'speaklater >= 1.3'
]

setup(
    name='palmer',
    description="Redice Flask Boost Library. Inspired by flask-api.",
    long_description=readme(),
    version='0.0.4',
    packages=get_packages(package),
    package_data=get_package_data(package),
    install_requires=install_requires,
    license='MIT License',
    author='Redice',
    author_email='devteam@redice-inc.com',
    url='https://github.com/redice-inc/palmer',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import datetime

from flask import request, Response
from flask._compat import text_type, string_types

from palmer.utils import http_statuses


class APIResponse(Response):

    def __init__(self, content=None, *args, **kwargs):
        super(APIResponse, self).__init__(None, *args, **kwargs)
        self.response_at = datetime.datetime.utcnow()

        media_type = None
        if isinstance(content, (list, dict, text_type, string_types)):
            renderer = request.renderer
            media_type = renderer.media_type
            if self.status_code == http_statuses.HTTP_204_NO_CONTENT:
                self.status_code = http_statuses.HTTP_200_OK
            content = self.get_cleaned_content(content)
            content = renderer.render(content, media_type)

        if content is None:
            content = []
        if isinstance(content, (text_type, bytes, bytearray)):
            self.set_data(content)
        else:
            self.response = content

        if media_type is not None:
            self.headers['Content-Type'] = str(media_type)

    def get_cleaned_content(self, content):
        return dict(
            request_at=request.request_at,
            response_at=self.response_at,
            status_code=self.status_code,
            result=content
        )



# -*- coding:utf-8 -*-
import os
import sys
from itertools import chain

import inflection
from flask import Flask, request
from flask._compat import text_type, string_types, reraise
from werkzeug.exceptions import HTTPException

from palmer.compat import is_flask_legacy
from palmer.config import APIConfig
from palmer.exceptions import APIException, NotFound, MethodNotAllowed
from palmer.request import APIRequest
from palmer.response import APIResponse
from palmer.session import ItsdangerousSessionInterface
from palmer.utils import http_statuses


class Palmer(Flask):
    request_class = APIRequest
    response_class = APIResponse

    #: Better Client-side sessions (http://flask.pocoo.org/snippets/51/)
    session_interface = ItsdangerousSessionInterface()

    def __init__(self, *args, **kwargs):
        super(Palmer, self).__init__(*args, **kwargs)
        self.api_config = APIConfig(self.config)

    def init_config(self, pkg, default='base', local='local', env_name='PALMER_ENV'):
        """
        :: Example

        app.config.base.py
        -------
        BaseConfig(object):
            pass

        app.config.stage.py
        --------
        StageConfig(BaseConfig):
            pass

        app.config.__init__.py
        ------

        app = Palmer(__name__)
        app.init_config('app.config')
        """

        env = os.environ.get(env_name, None)
        name = env or default
        try:
            obj = __import__('{0}.{1}'.format(pkg, name), globals(), locals(), ['object'], 0)
            config = getattr(obj, '{0}Config'.format(inflection.camelize(name)))
        except (ImportError, KeyError, TypeError) as e:
            raise e
        self.config.from_object(config)
        self.config['CONFIG_NAME'] = name

        if not env:
            try:
                obj = __import__('{0}.{1}'.format(pkg, local), globals(), locals(), ['object'], 0)
                config = getattr(obj, '{0}Config'.format(inflection.camelize(local)))
            except (ImportError, KeyError) as e:
                pass
            else:
                self.config.from_object(config)
        self.api_config = APIConfig(self.config)

    def init_controllers(self):
        pass

    def make_response(self, rv):
        """
        We override this so that we can additionally handle
        list and dict types by default.
        """
        # super(Palmer, self).make_response()
        status_or_headers = headers = None
        if isinstance(rv, tuple):
            rv, status_or_headers, headers = rv + (None,) * (3 - len(rv))

        if rv is None and status_or_headers:
            raise ValueError('View function did not return a response')

        if isinstance(status_or_headers, (dict, list)):
            headers, status_or_headers = status_or_headers, None

        if not isinstance(rv, self.response_class):
            if isinstance(rv, (text_type, bytes, bytearray, list, dict)):
                rv = self.response_class(rv, headers=headers, status=status_or_headers)
                headers = status_or_headers = None
            else:
                rv = self.response_class.force_type(rv, request.environ)

        if status_or_headers is not None:
            if isinstance(status_or_headers, string_types):
                rv.status = status_or_headers
            else:
                rv.status_code = status_or_headers
        if headers:
            rv.headers.extend(headers)
        return rv

    def handle_user_exception(self, e):
        """
        We override the default behavior in order to deal with APIException.
        """
        exc_type, exc_value, tb = sys.exc_info()
        assert exc_value is e

        if isinstance(e, HTTPException) and not self.trap_http_exception(e):
            return self.handle_http_exception(e)

        if isinstance(e, APIException):
            return self.handle_api_exception(e)

        blueprint_handlers = ()
        handlers = self.error_handler_spec.get(request.blueprint)
        if handlers is not None:
            blueprint_handlers = handlers.get(None, ())
        app_handlers = self.error_handler_spec[None].get(None, ())
        if is_flask_legacy():
            for typecheck, handler in chain(blueprint_handlers, app_handlers):
                if isinstance(e, typecheck):
                    return handler(e)
        else:
            for typecheck, handler in chain(dict(blueprint_handlers).items(),
                                            dict(app_handlers).items()):
                if isinstance(e, typecheck):
                    return handler(e)

        debug = self.config.get('DEBUG', False)
        testing = self.config.get('TESTING', False)
        if debug or testing:
            return reraise(exc_type, exc_value, tb)
        return self.handle_api_exception(APIException())

    def handle_http_exception(self, e):
        description = getattr(e, 'description', None)
        status_code = getattr(e, 'code', None)
        exc_cls = APIException
        if status_code == http_statuses.HTTP_404_NOT_FOUND:
            exc_cls = NotFound
        elif status_code == http_statuses.HTTP_405_METHOD_NOT_ALLOWED:
            exc_cls = MethodNotAllowed
        return self.handle_api_exception(exc_cls(message=description, status_code=status_code))

    def handle_api_exception(self, exc):
        return self.response_class(
            dict(message=exc.message, errors=exc.errors, code=exc.__class__.__name__),
            status=exc.status_code,
        )


